# 一、实验内容
## 1.获取进程运行状态
### 1.1. 1 实验步骤
首先更新到分支scheduler下，每次新增应用时都需要在kernel下的syscall.h与syscall.c中加入对应的函数<br>
在syscall.h头文件中增加:<br> 
#define SYS_wait_sched 25 #define SYS_set_priority 26<br>
在syscall.c代码文件中增加:<br> 
extern uint64 sys_wait_sched(void); extern uint64 sys_set_priority(void);<br>
[SYS_wait_sched]  sys_wait_sched, [SYS_set_priority] sys_set_priority,<br>
[SYS_wait_sched] "wait_sched",[SYS_set_priority] "_set_priority",<br>
最后在Makefile中增加:<br>
$U/_stat \  $U/_priostat \<br>
确保应用被正确编译至用户级<br>

根据提示，我们需要修改的主体代码部分在sysproc.c与proc.c中<br>
首先我们根据TODO标记补全了proc.c需要添加on_state_change的位置，如图所示<br>
<img src="./initial.png" alt="初始化" width="600"/><br>
根据示意图还需要在userinit中加入on_state_change(UNUSED, RUNNABLE, p); <br>
在fork中加入on_state_change(UNUSED, RUNNABLE, np);<br>
在exit中加入on_state_change(RUNNING, ZOMBIE, p);<br>
在yield中加入on_state_change(RUNNING, RUNNABLE, p); <br>
在sleep中加入on_state_change(RUNNABLE, SLEEPING, p); <br>
在wakeup中加入on_state_change(SLEEPING, RUNNABLE, p);<br>
在kill中加入on_state_change(SLEEPING, RUNNABLE, p);<br>
特别的，在scheduler中需要完善调度器的代码，如图所示<br>
<img src="./scheduler.png" alt="调度" width="600"/><br>
(注意上述工作也同时完成了优先级调度的相关内容，后续将不再赘述)<br>
接下来我们只需要实现具体的int wait_sched与int on_state_change函数即可<br>

首先修改sysproc.c中的sys_wait_sched函数，代码思路如下:<br>

定义变量：
函数一开始定义了三个 uint64 类型的变量 runable_time_addr、running_time_addr 和 sleep_time_addr，用于接收用户传递的指针地址。
定义三个 int* 类型的指针 runable_time、running_time 和 sleep_time，用于指向这些地址。

获取用户传递的参数：
使用 argaddr 函数依次获取用户传递的地址参数。argaddr 是 XV6 系统提供的一个内核函数，用于获取系统调用的参数地址。
如果获取失败（返回值为负），说明参数有误，直接返回 -1 以指示错误。

地址转换为指针：
将获取到的地址参数转换为指针，以便在后续代码中使用。

最后，调用内核态的 wait_sched 函数，将三个指针传递给它，并返回其执行结果。<br>
代码实现如图所示<br>
<img src="./sysproc1.png" alt="wait" width="600"/><br>

再实现proc.c中的wait_sched与on_state_change函数，代码思路如下:<br>

变量定义：
定义指针 np 用于遍历所有进程结构体。
定义 havekids 用于标识是否有子进程，pid 用于存储子进程 ID。
定义 p 并指向当前调用进程。

获取父进程锁：
为了确保操作的安全性，首先获取父进程 p 的锁，防止在等待子进程退出时出现竞争条件。

主循环：
进入无限循环，开始查找子进程并判断其状态。
将 havekids 设为 0，以便在循环中标识是否找到子进程。
遍历 proc 数组（所有进程），查找当前进程 p 的子进程 np。

子进程处理：
对于每一个子进程 np：
获取 np 的锁，保证对该子进程的访问是原子的。
检查 np 的状态是否为 ZOMBIE（僵尸状态），表示该子进程已经退出。
拷贝数据至用户态：
若子进程是僵尸状态，通过 copyout 将子进程的 runable_time、running_time 和 sleep_time 拷贝至用户传入的指针。
copyout 函数用于从内核态安全地将数据拷贝到用户态。如果拷贝失败，释放锁并返回错误。
释放子进程资源并返回 PID：
使用 freeproc 释放子进程资源。
释放 np 和 p 的锁，返回子进程的 PID 以表明等待成功。

等待子进程：
如果遍历所有进程后 havekids 为 0，表示当前进程没有子进程，或进程 p 被终止，释放锁并返回 -1。
如果有子进程但还未退出，调用 sleep 函数将当前进程睡眠在 p->lock 上，等待子进程状态发生变化（例如变为 ZOMBIE）。<br>
代码实现如图所示<br>
<img src="./proc1.png" alt="wait" width="600"/><br>

### 1.1.2 实验测试结果
<img src="./stat_result.png" alt="任务一" width="600"/><br>

PID 	Runnable Time (ticks)	Running Time (ticks)	Sleep Time (ticks)<br>
4 $~~~~~~~~~~~$ 119 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 172 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 12<br>
7 $~~~~~~~~~~~$ 113 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 180 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 12<br>
5 $~~~~~~~~~~~$ 124 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 170 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 12<br>
8 $~~~~~~~~~~~$ 115 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 180 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 12<br>
6 $~~~~~~~~~~~$ 116 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 181 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 12<br>
Average Turnaround Time 306_ticks<br>
## 2.优先级调度
### 1.2. 1 实验步骤
在配置文件的步骤与第一部分中的内容是相近的，以下仅对不同的内容加以阐述<br>
事实上，我们只需填充proc.c中的set_priority函数与sysproc.c中的sys_set_priority函数即可<br>

首先修改sysproc.c中的sys_set_priority函数，代码思路如下:<br>

定义变量：
定义两个 int 类型变量 priority 和 pid，分别用于存储进程的优先级和进程 ID（PID）。

获取并验证系统调用参数：
使用 argint 函数依次获取用户传递的参数，argint(0, &priority) 表示获取优先级参数，argint(1, &pid) 表示获取进程 ID 参数。
如果参数获取失败（即 argint 返回负数），返回 -1 表示错误。

调用 set_priority 函数：
传递参数 priority 和 pid 给 set_priority 函数，执行优先级设置的实际操作，并返回其执行结果。<br>
代码实现如图所示<br>
<img src="./sysproc2.png" alt="priority" width="600"/><br>

再实现proc.c中set_priority函数，代码思路如下:<br>

检查优先级范围：
检查 priority 参数是否在合法范围内（假设范围为 0 到 3），如果不在范围内，返回 -1 表示无效优先级。

遍历所有进程：
使用循环遍历 proc 数组中的每个进程结构体 p。 proc 是一个包含所有进程的数组，大小为 NPROC。

进程匹配：
获取进程 p 的锁，确保操作的原子性，防止其他线程对该进程的数据进行并发修改。
检查 p->pid 是否与目标 pid 匹配。如果匹配，说明找到了目标进程。

设置优先级并释放锁：
将 p->priority 更新为新优先级 priority，完成优先级设置。
释放锁，防止死锁或竞争条件，并返回 0 表示成功。

进程未找到或锁释放：
若进程 p 的 pid 不匹配目标 pid，则释放锁继续检查下一个进程。
如果遍历完所有进程后仍未找到匹配的 pid，返回 -1 表示未找到目标进程。<br>
代码实现如图所示<br>
<img src="./proc2.png" alt="priority" width="600"/><br>
值得注意的是，在我们使用优先级调度策略的时候需要在proc.h头文件中对define处做出修改，注释掉rr并启用pr<br>

### 1.2.2 实验测试结果
<img src="./priostat_result.png" alt="任务二" width="600"/><br>
PID 	Priority    Runnable Time (ticks)	Running Time (ticks)	Sleep Time (ticks)<br>
4 $~~~~~~~$ 0 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 0 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 174 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 12<br>
7 $~~~~~~~$ 3 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 18 $~~~~~~~~~~~~~~~~~~~~~~~~$ 174 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 12<br>
5 $~~~~~~~$ 1 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 18 $~~~~~~~~~~~~~~~~~~~~~~~~$ 179 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 12<br>
8 $~~~~~~~$ 0 $~~~~~~~~~~~~~~~~~~~~~~~~$ 156 $~~~~~~~~~~~~~~~~~~~~~~~$ 149 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 12<br>
6 $~~~~~~~$ 2 $~~~~~~~~~~~~~~~~~~~~~~~~$ 200 $~~~~~~~~~~~~~~~~~~~~~~~$ 129 $~~~~~~~~~~~~~~~~~~~~~~~~~$ 12<br>
Average Turnaround Time：251_ticks
